import typing

from apistar import http, types, validators
from apistar.exceptions import NotFound
from apistar_pagination import PageNumberResponse
from sqlalchemy.orm import Session

from apistar_crud.resource.base import BaseResource


class Resource(BaseResource):
    @classmethod
    def add_create(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        def create(cls, session: Session, element: input_type) -> output_type:
            """
            Create a new element for this resource.
            """
            record = model(**element)
            session.add(record)
            session.flush()
            return http.JSONResponse(output_type(record), status_code=201)

        return {"_create": classmethod(create)}

    @classmethod
    def add_retrieve(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        def retrieve(cls, session: Session, element_id: str) -> output_type:
            """
            Retrieve an element of this resource.
            """
            record = session.query(model).get(element_id)
            if record is None:
                raise NotFound

            return output_type(record)

        return {"_retrieve": classmethod(retrieve)}

    @classmethod
    def add_update(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        def update(cls, session: Session, element_id: str, element: input_type) -> output_type:
            """
            Update an element of this resource.
            """
            record = session.query(model).get(element_id)
            if record is None:
                raise NotFound

            for k, value in element.items():
                setattr(record, k, value)

            return http.JSONResponse(output_type(record), status_code=200)

        return {"_update": classmethod(update)}

    @classmethod
    def add_delete(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        def delete(cls, session: Session, element_id: str) -> typing.Dict[str, typing.Any]:
            """
            Delete an element of this resource.
            """
            if session.query(model).filter_by(id=element_id).count() == 0:
                raise NotFound

            session.query(model).filter_by(id=element_id).delete()
            return http.JSONResponse(None, status_code=204)

        return {"_delete": classmethod(delete)}

    @classmethod
    def add_list(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        def filter_(cls, session: Session, **filters) -> typing.List[output_type]:
            """
            Filter resource collection.
            """
            filters = {k: v for k, v in filters.items() if v}
            if filters:
                queryset = session.query(model).filter_by(**filters)
            else:
                queryset = session.query(model).all()

            return [output_type(record) for record in queryset]

        def list_(cls, page: http.QueryParam = None, page_size: http.QueryParam = None) -> typing.List[output_type]:
            """
            List resource collection.
            """
            return PageNumberResponse(page=page, page_size=page_size, content=cls._filter())  # noqa

        return {"_list": classmethod(list_), "_filter": classmethod(filter_)}

    @classmethod
    def add_drop(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        class DropOutput(types.Type):
            deleted = validators.Integer(title="deleted", description="Number of deleted elements", minimum=0)

        def drop(cls, session: Session) -> DropOutput:
            """
            Drop resource collection.
            """
            num_records = session.query(model).count()
            session.query(model).delete()
            return http.JSONResponse(DropOutput({"deleted": num_records}), status_code=204)

        return {"_drop": classmethod(drop)}



import typing

from apistar import App, Route, types, validators


class Metadata(types.Type):
    resources = validators.Object(title="resources", description="Resource list")
    admin = validators.String(title="admin", description="Admin URL")
    schema = validators.String(title="schema", description="OpenAPI schema URL")


class Admin:
    def __init__(self, *resources):
        self.resources = {resource.name: resource for resource in resources}

    def main(self, app: App, path: str = ""):
        """
        Admin main page presenting a list of resources.
        """
        return app.render_template("apistar_crud/admin.html")

    def metadata(self, app: App) -> Metadata:
        """
        Admin metadata.
        """
        return Metadata(
            {
                "resources": {
                    resource.name: {
                        "name": resource.name,
                        "verbose_name": resource.verbose_name,
                        "columns": resource.columns,
                        "order": resource.order,
                    }
                    for resource in self.resources.values()
                },
                "admin": app.reverse_url("admin:main"),
                "schema": app.reverse_url("serve_schema"),
            }
        )

    @property
    def routes(self) -> typing.List[Route]:
        return [
            Route("/", "GET", self.main, name="main", documented=False),
            Route("/{+path}", "GET", self.main, name="main-path", documented=False),
        ]

    @property
    def metadata_routes(self) -> typing.List[Route]:
        return [Route("/metadata/", "GET", self.metadata, name="metadata", documented=False)]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['apistar_crud', 'apistar_crud.resource']

package_data = \
{'': ['*'], 'apistar_crud': ['static/*', 'templates/*']}

install_requires = \
['apistar-pagination>=0.3.0,<0.4.0', 'apistar>=0.5.30,<0.6.0']

setup_kwargs = {
    'name': 'apistar-crud',
    'version': '0.6.1',
    'description': 'API Star tools to create CRUD resources.',
    'long_description': '# API Star CRUD\n[![Build Status](https://travis-ci.org/PeRDy/apistar-crud.svg?branch=master)](https://travis-ci.org/PeRDy/apistar-crud)\n[![codecov](https://codecov.io/gh/PeRDy/apistar-crud/branch/master/graph/badge.svg)](https://codecov.io/gh/PeRDy/apistar-crud)\n[![PyPI version](https://badge.fury.io/py/apistar-crud.svg)](https://badge.fury.io/py/apistar-crud)\n\n* **Version:** 0.6.1\n* **Status:** Production/Stable\n* **Author:** José Antonio Perdiguero López\n\n## Features\nAPI Star CRUD provides an easy way to define a REST resource and generic operations over it.\n\n### Resource\nThe resources are classes with a default implementation for **methods**:\n* `create`: Create a new element for this resource.\n* `retrieve`: Retrieve an element of this resource.\n* `update`: Update (partially or fully) an element of this resource.\n* `delete`: Delete an element of this resource.\n* `list`: List resource collection with **filtering** capabilities.\n* `drop`: Drop resource collection.\n\n### ORM\nAPI Star CRUD supports the following ORM:\n* [SQLAlchemy](https://www.sqlalchemy.org/) through [apistar-sqlalchemy](https://github.com/PeRDy/apistar-sqlalchemy).\n* [Peewee](https://github.com/coleifer/peewee) through [apistar-peewee-orm](https://github.com/PeRDy/apistar-peewee-orm).\n\n### Admin site\nAPI Star CRUD serves an admin site to handle resources in a graphical way, by default this site is routed to `/admin/`. \n\n## Quick start\nInstall API Star CRUD:\n\n```bash\n$ pip install apistar-crud[peewee]\n```\n\nor \n\n```\n$ pip install apistar-crud[sqlalchemy]\n```\n\nFollow the steps:\n\n1. Create an **input type** and **output type** for your resource:\n2. Define a **model** based on your ORM.\n3. Build your **resource** using the metaclass specific for your ORM.\n4. Add the **routes** for your resources.\n\n## Usage\n### SQLAlchemy\nExample of a fully functional resource based on SQLAlchemy.\n\nCreate an **input type** and **output type**:\n\n```python\nclass PuppyInputType(types.Type):\n    name = validators.String()\n\nclass PuppyOutputType(types.Type):\n    id = validators.Integer()\n    name = validators.String()\n```\n\nDefine a **model**:\n\n```python\nclass PuppyModel(Base):\n    __tablename__ = "Puppy"\n\n    id = Column(Integer, primary_key=True)\n    name = Column(String)\n```\n\nThe **resource**:\n\n```python\nfrom apistar_crud.resource.sqlalchemy import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n```\n\nThe resource generates his own **routes**:\n\n```python\nfrom apistar import App\nfrom apistar_crud.routes import routes as resource_routes\n\nresource_routes.register(PuppyResource, "/puppy/", admin=True)\n\nroutes = [\n    # ... your app routes\n]\n\nroutes += resource_routes.routes(admin="/admin/")\npackages = ("apistar", "apistar_crud")\napp = App(routes=routes, packages=packages)\n```\n\n### Peewee\nExample of a fully functional resource based on Peewee.\n\nCreate an **input type** and **output type**:\n\n```python\nclass PuppyInputType(types.Type):\n    name = validators.String()\n\nclass PuppyOutputType(types.Type):\n    id = validators.Integer()\n    name = validators.String()\n```\n\nDefine a **model**:\n\n```python\nclass PuppyModel(peewee.Model):\n    name = peewee.CharField()\n```\n\nThe **resource**:\n\n```python\nfrom apistar_crud.resource.peewee import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n```\n\nThe resource generates his own **routes**:\n\n```python\nfrom apistar import App\nfrom apistar_crud.routes import routes as resource_routes\n\nresource_routes.register(PuppyResource, "/puppy/", admin=True)\n\nroutes = [\n    # ... your app routes\n]\n\nroutes += resource_routes.routes(admin="/admin/")\npackages = ("apistar", "apistar_crud")\napp = App(routes=routes, packages=packages)\n```\n\n## Resources\n\n### Routes\nThe routes for resource methods are:\n\n|Method  |Verb  |URL\n|--------|------|--------------\n|create  |POST  |/\n|retrieve|GET   |/{element_id}/\n|update  |PUT   |/{element_id}/\n|delete  |DELETE|/{element_id}/\n|list    |GET   |/\n|drop    |DELETE|/\n\n\n### Override methods\nTo customize CRUD methods you can override them like:\n\n```python\nfrom apistar_crud.resource.peewee import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n    \n    @classmethod\n    def create(cls, element: PuppyInputType) -> PuppyOutputType:\n        # Do your custom process\n```\n\n### Filtering\nDefault `list` implementation has filtering capabilities but to reflect it in the schema is necessary to override that \nmethod defining all the parameters that will be used as a filter and pass it as keywords to `_list()` method. \n\nIt is a direct translation to ORM keyword arguments \nso in case of SQLAlchemy it uses a `filter_by()` method and a `where()` in Peewee case.\n\n```python\nimport typing\n\nfrom apistar import http\nfrom apistar_crud.resource.peewee import Resource\nfrom apistar_pagination import PageNumberResponse\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n    \n    @classmethod\n    def list(cls, name: http.QueryParam, page: http.QueryParam=None, page_size: http.QueryParam=None) -> typing.List[PuppyOutputType]:\n        return PageNumberResponse(page=page, page_size=page_size, content=cls._filter(name=name))\n```\n',
    'author': 'José Antonio Perdiguero López',
    'author_email': 'perdy@perdy.io',
    'url': 'https://github.com/PeRDy/apistar-crud',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
